import logging
import os
import tempfile
from typing import List

import numpy as np
import vtk
from vtkmodules.util.numpy_support import vtk_to_numpy, numpy_to_vtk

from MeshObject.VtkWrapping import Points, Cells, Types, Type, IdList, Cell, CellArray
from MeshObject.decorators import (
    deprecated,
    unstructured_grid_method,
    vtkFilter,
    polydata_method,
    timer,
)
from MeshObject.mmg_options import MMGOptions
from MeshObject.prefabs import (
    BuildTransform,
    Sphere,
    Cylinder,
    Cube,
    Ellipsoid,
    Superquadric,
)
from MeshObject.vtk_wrap import Reader, Writer

try:
    from cardiac_utils.mesh import SpatialData

    HAS_CARDIAC_UTILS = True
except ImportError:
    HAS_CARDIAC_UTILS = False


class Object:
    """
    Main vtk interface for object inheriting from vtkDataSet.

    Easy to read/write, apply filters, access data, etc..
    """

    types = Types

    def __init__(self, data, **kwargs):
        """
        See `Object.load` to instanciate a <gaetools.meshing.Object>.

        :param vtk.vtkDataSet data: Internal vtk data
        """
        self.data = data
        self.filename = kwargs.get("filename", None)

        ### Built later if necessary
        self.cellsAroundPoints = None
        self._pointIdsAroundPoint = None
        self.pointLocator = None
        self.tetras_around_points = None

    def __getattribute__(self, name):
        try:
            getattr(vtk, f"vtk{name}")
        except AttributeError:
            return super().__getattribute__(name)
            # log.critical(f"Filter vtk{name} could not be found in vtk's attributes or an error occured calling cellTypes method/property")
            # log.debug(f"Object: \n{self}")
            # raise AttributeError(f"Object has no attribute {name!r}")

        else:

            def wrapper(**kwargs):
                return self.apply_filter(name, **kwargs)

            return wrapper

    def copy(self):
        """ Deep copy of vtkDataSet but shallow copy of attributes
        """
        data = type(self.data)()
        data.DeepCopy(self.data)
        obj = type(self)(data)
        obj.__dict__.update(self.__dict__)
        return obj

    @property
    def type(self):
        """ Return the type of the internal vtkDataSet
        """
        return type(self.data)

    @classmethod
    def read(cls, filename, **kwargs):
        """
        Read vtkDataSet from file.

        :param str filename: File to read
        :param str type: Type of `vtkReader`, default is `generic`

        .. note::
            You can specify a reader format with type:

            - XMLPolyData | xmlpd
            - PolyData | pd
            - UnstructuredGrid | ugrid
            - GenericDataObject | generic

            eg:

            - obj.read('surf.vtp', type='XMLPolyData')
            - obj.read('vol.vtk', type='ugrid')
        """
        return cls(Reader(filename, **kwargs).data, filename=filename, **kwargs)

    @classmethod
    def load(cls, data, **kwargs):
        """
        Main method to initiliaze a vtk object.

        :param data: Multiple types supported: Object, vtkDataSet, str(path), SpatialData, vtkAlgorithm
        """
        if type(data) is str:
            return cls.read(data, **kwargs)
        elif isinstance(data, vtk.vtkObject):
            if isinstance(data, vtk.vtkDataSet):
                return cls(data, **kwargs)
            else:
                return cls(data.GetOutput())
        elif isinstance(data, cls):
            return data
        elif HAS_CARDIAC_UTILS and isinstance(data, SpatialData):
            return cls(data.vtk_polydata)
        else:
            raise RuntimeError(f"Unexpected type {data!r}")

    @classmethod
    def new(cls, dtype):
        """
        Create instance with empty vtk object.

        :param dtype: vtk class or class name of the wrapped vtk object
        """
        if type(dtype) is str:
            dtype = dtype.lower()
            if dtype == "pd" or dtype == "polydata":
                return cls(vtk.vtkPolyData())
            elif dtype == "ug" or dtype == "unstructuredgrid":
                return cls(vtk.vtkUnstructuredGrid())
            else:
                msg = f"<Object.new> Unexpected type {dtype!r}, choose between pd|ug"
                log.error(msg)
                raise RuntimeError(msg)
        elif isinstance(dtype, vtk.vtkObject):
            return cls(dtype())
        else:
            msg = f"<Object.new> Unexpected type {dtype!r}"
            log.error(msg)
            raise RuntimeError(msg)

    @classmethod
    def from_elements(cls, vertices, faces=None, dtype="pd", point_arrays=None):
        """
        Create instance of <Object> with initialized vtk data.

        :param dtype: vtk class or class name of the wrapped vtk object
        :param vertices: Array of vertices topology
        :param faces: Array of faces topology
        """
        obj = Object.new(dtype)
        obj.points = vertices  # from setter
        if faces is not None:
            obj.cells = faces  # from setter
        if point_arrays is not None:
            for name, arr in point_arrays.items():
                obj.addPointData(arr, name)
        return obj

    @property
    def elems(self):
        return self.points, self.cells.to_numpy()

    @classmethod
    def Sphere(cls, **kwargs):
        return cls(Sphere(**kwargs))

    @classmethod
    def Cube(cls, **kwargs):
        # SetBounds, SetCenter
        return cls(Cube(**kwargs))

    @classmethod
    def Superquadric(cls, **kwargs):
        return cls(Superquadric(**kwargs))

    @classmethod
    def Cylinder(cls, **kwargs):
        quad_cylinder = cls(Cylinder(**kwargs))
        quad_cylinder.TriangleFilter()
        quad_cylinder.clean()
        return quad_cylinder

    @classmethod
    def Ellipsoid(cls, radiuses, center=None):
        ell = cls(Ellipsoid(radiuses))
        if center is not None:
            ell.transform(translate=center)
        return ell

    def write(self, *args, support_new_type=True, change_vtk_version=False, **kwargs):
        """
        Write vtkDataSet to file.

        :param str filename: Filename to write, positional argument
        :param str type: Type of `vtkWriter`, default is `generic`

        .. note::
            You can specify a writer format with type:

            - XMLPolyData | xmlpd
            - PolyData | pd
            - UnstructuredGrid | ugrid
            - GenericDataObject | generic

            eg:

            - obj.write('surf.vtp', type='XMLPolyData')
            - obj.write('vol.vtk', type='ugrid')
        """
        if len(args) == 0:
            if self.filename is not None:
                args = [self.filename]
            else:
                log.error("Give a filename")
        Writer(*args, **kwargs, data=self.data)

        # vtk DataFile Version 5.0 generates OFFSETS and CONNECTIVITY
        # support_new_type set to False will convert to the old-style file type, ie
        # CELLS   numPoints0,i,j,k,l,...
        if not support_new_type:
            self.replace_polygons(args[0])
        if change_vtk_version:
            self.change_vtk_version(args[0], "5.0")
        # print(args)
        # breakpoint()

    def __repr__(self):
        fname = "" if self.filename is None else f" - {os.path.basename(self.filename)}"
        s = f"\n** {self.__class__.__name__}<{type(self.data).__name__}>{fname} **\n"
        s += f" - {self.nbPoints} points\n"
        s += f" - {self.nbCells} cells\n"
        cellTypes = np.rint(self.cellTypes)
        for t in np.unique(cellTypes):
            nbCells = len(np.where(cellTypes == t)[0])
            name = self.types.find(t).name
            s += f"   - {nbCells} {name}s\n"
        if len(self.pointDataNames) > 0:
            s += f" - PointData: {', '.join([f'{n!r}' for n in self.pointDataNames])}\n"
        if len(self.cellDataNames) > 0:
            s += f" - CellData: {', '.join([f'{n!r}' for n in self.cellDataNames])}\n"
        return s

    @deprecated("Use <Object.points>")
    @property
    def old_points(self):
        """ Get all points
        """
        return vtk_to_numpy(self.data.GetPoints().GetData())

    @property
    def points(self):
        """ Get all points
        """
        return Points(self.data.GetPoints())

    @points.setter
    def points(self, array):
        """ Set all points
        """
        pts = vtk.vtkPoints()
        for i in range(len(array)):
            pts.InsertNextPoint(array[i])
        self.data.SetPoints(pts)

    @property
    def bounds(self):
        return self.data.GetBounds()

    @property
    def cells(self):
        """ Get all cells
        """
        return Cells(self.data)

    def addCells(self, arr):
        for ids in arr:
            dtype = self.types.from_nb_points(len(ids)).vtk_type
            self.data.InsertNextCell(dtype, IdList.from_np(ids))

    # # @property
    # def numpy_cells(self):
    #     """ Get all cells in a contiguous numpy array
    #     """
    #     return vtk_to_numpy(self.data.GetCells())
    #     # return vtk_to_numpy(self.data.GetPolys())

    @cells.setter
    def cells(self, cells):
        if isinstance(cells, vtk.vtkCellArray):
            # self.data.SetCells(cells)
            new_cells = CellArray(cells)
        elif type(cells) is Cells:
            # self.data.SetCells(cells.data)
            new_cells = cells
        elif isinstance(cells, np.ndarray):
            new_cells = Cells.from_array(cells)
        else:
            raise

        method = Object.getCells_Method(self.data, p="Set", s="s")
        if self.type is vtk.vtkPolyData:
            method(new_cells)
        else:
            # breakpoint()
            method(new_cells, new_cells.types)

    @staticmethod
    def getCells_Method(vtk_obj, p="", s=""):
        method = "Poly" if type(vtk_obj) is vtk.vtkPolyData else "Cell"
        method = f"{p}{method}{s}"
        return getattr(vtk_obj, method)

    @property
    @deprecated("Replaced by 'Object.cells', using class <vtk_wrap.Cells>")
    def np_cells(self):
        """ Get all cells
        """
        return vtk_to_numpy(self.data.GetCells().GetData())

    @property
    def cellTypes(self):
        """
        Return an array of size the number of cells, containing the type of each cell.
        """
        if self.type is vtk.vtkPolyData:
            return [c.type.vtk_type for c in self.cells]
        else:
            return vtk_to_numpy(self.data.GetCellTypesArray())

    @property
    def nbPoints(self):
        """ Get the number of points
        """
        return self.data.GetNumberOfPoints()

    @property
    def nbCells(self):
        """ Get the number of cells
        """
        return self.data.GetNumberOfCells()

    def interpolate(
        self,
        source,
        fname=None,
        volume=True,
        inside_val=0,
        include=None,
        exclude=None,
        kernel=None,
        dtype=np.float64,
    ):
        """
        Function intended to interpolate PointData between two data sets. `self` will be the
        geometry template; arrays from `source` will be interpolated on this geometry.

        :param source: vtkDataSet object-loadable (see `Object.load`), vtkDataSet with point data arrays to copy
        :param str fname: Output vtkDataSet filename, vtkDataSet with the newly interpolated arrays
        :param bool volume: If true will interpolate pointdata to all points (including inside the surface), False see inside_val
        :param float inside_val: If volume is False, will replace inside pointdata value by inside_val
        :param list include: PointData array names to copy (if None will pass all PointData arrays)
        :param list exclude: PointData array names to not copy
        :param dict kernel: Parameters for the interpolation kernel [vtkGaussianKernel|vtkVoronoiKernel|vtkShepardKernel|vtkSPHKernel].
           See https://vtk.org/doc/nightly/html/classvtkInterpolationKernel.html.
           Possible methods for `vtkGaussianKernel` are `SetKernelFootprintToNClosest`, `SetNumberOfPoints`
        :param np.dtype dtype: numpy type of the PointData array
        """
        if exclude is None:
            exclude = ["Normals"]
        surf_object = self.copy()
        # Remove all cells not on the boundary, but do not clean, ie leaving ghost points
        if not volume:
            surf_object.GeometryFilter()

        if kernel is None:
            kernel = dict(
                name="vtkGaussianKernel",
                SetSharpness=4,
                SetKernelFootprintToNClosest=[],
                SetNumberOfPoints=20,
            )
        else:
            kernel = dict(kernel)

        vtkKernel = getattr(vtk, kernel.pop("name"))()
        for k, v in kernel.items():
            if not isinstance(v, list):
                v = [v]
            getattr(vtkKernel, k)(*v)

        # Transport point data to the new mesh points
        surf_object.PointInterpolator(
            SetSourceData=Object.load(source).data, SetKernel=vtkKernel
        )

        # Since polydata is not cleaned, it still contains ghosts points
        # unconnected to cells. All points on the surface have at least one
        # adjacent cell, others are marked as True in inside_points
        inside_points = np.zeros(surf_object.nbPoints, dtype=bool)
        if not volume:
            for i in range(surf_object.nbPoints):
                cells = vtk.vtkIdList()
                surf_object.data.GetPointCells(i, cells)
                if cells.GetNumberOfIds() == 0:
                    inside_points[i] = True

        # Actually pass the pointData arrays
        if include is None:
            include = surf_object.pointDataNames

        for name in include:
            orig_arr = surf_object.getPointDataArray(name)
            if name not in exclude:
                orig_arr[inside_points] = inside_val
                orig_arr = orig_arr.astype(dtype)
                log.debug(f"Adding point data {name!r}")
                self.addPointData(orig_arr, name)

        # Write to file
        if fname is not None:
            self.write(fname)

    @deprecated("Only works on polydata")
    def deleteCells(self, dtype=None, indices=None):
        if dtype is not None:
            indices = self.cellTypes[np.where(self.cellTypes) == dtype]
        elif indices is not None:
            pass
        else:
            raise TypeError

        for i in indices:
            self.data.DeleteCells(i)

        self.data.RemoveDeletedCells()
        log.debug("Cells deleted")

    @unstructured_grid_method
    def filterCellsTo(self, idType):
        """
        @unstructured_grid_method
        Keep only the cells of type `idType`

        :param int idType: vtk cell type, see `the cell types <https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html>`_
        """
        if isinstance(idType, Type):
            idType = idType.vtk_type

        cells, types = vtk.vtkCellArray(), self.cellTypes
        ids_to_keep = []
        for i, type in enumerate(types):
            if type == idType:
                cells.InsertNextCell(self.data.GetCell(i))
                ids_to_keep.append(i)
        self.data.SetCells([idType] * cells.GetNumberOfCells(), cells)

        for cell_data_name in self.cellDataNames:
            array = self.getCellDataArray(cell_data_name)
            self.addCellData(array[ids_to_keep], cell_data_name)

    @unstructured_grid_method
    def removeCells(self, idType):
        """
        Remove all cells having type `idType`

        :param int idType: vtk cell type, see `the cell types <https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html>`_
        """
        cells, types, myTypes = vtk.vtkCellArray(), [], self.cellTypes
        for i, type in enumerate(myTypes):
            if type != idType:
                types.append(type)
                cells.InsertNextCell(self.data.GetCell(i))
        self.data.SetCells(types, cells)

    def cellsType(self, idType):
        """
        Return the id of all the cells of type `idType`

        :param int idType: vtk cell type, see `the cell types <https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html>`_
        """
        return np.where(self.cellTypes == idType)

    # @property
    # @deprecated("")
    # def triangles(self):
    #     return self.np_cells[self.cellsType(vtk.VTK_TRIANGLE)]

    # @property
    # @deprecated("")
    # def tetras(self):
    #     return self.np_cells[self.cellsType(vtk.VTK_TETRA)]

    @property
    @unstructured_grid_method
    def triangles(self):
        return CellArray().from_array(
            self.cells.as_np()[self.cellsType(vtk.VTK_TRIANGLE)]
        )

    @property
    @unstructured_grid_method
    def tetras(self):
        tet = []
        for cell in self.cells:
            if cell.type is Object.types.TETRA:
                tet.append(cell)
        return tet

    @deprecated("")
    @unstructured_grid_method
    def addCell(self, *points_id, **arrays_name):
        """
        Adds a new cell to the mesh defined by `points_id`.
        Cell data is set to 0 unless it is specified as kwargs.

        Example: add a triangle defined by points 50, 51, 52 and with value in array `array_name` equals to 15.23
        ```python
        mesh.addCell(50, 51, 52, array_name=15.23)
        ```

        ..warning:
            Does not apply to polydata!

        :param points_id: Point ids of the new cell
        :param arrays_name: Dict of array name and value
        """
        dtype = self.types.from_nb_points(len(points_id))
        type = dtype.vtk_type
        cell = dtype.new()

        # for i, point_id in enumerate(points_id):
        #     cell[i] = point_id
        cell = Cell(ids=points_id)

        cells = self.data.GetCells()
        cells.InsertNextCell(cell)

        cells_type = self.cellTypes
        cells_type = np.append(cells_type, type)

        if cells_type.size == 1:
            cells_type = cells_type[0]

        self.data.SetCells(cells_type, cells)

        for name in self.cellDataNames:
            arr = self.getCellDataArray(name)
            shp = arr.shape[-1] if arr.ndim > 1 else 1
            arr = np.append(arr, arrays_name.pop(name, np.zeros(shp)))
            self.removeCellData(name)
            self.addCellData(arr, name)

        for k, v in arrays_name.items():
            log.warning(f"Value {v} for array named {k!r} was not used")

    def addPointData(self, data, name):
        """
        Add an array to the PointData arrays

        :param np.array data: Array of size the number of points
        :param str name: Name of the array
        """
        array = numpy_to_vtk(data)
        array.SetName(name)
        self.data.GetPointData().AddArray(array)

    def addCellData(self, data, name):
        """
        Add an array to the CellData arrays

        :param np.array data: Array of size the number of cells
        :param str name: Name of the array
        """
        array = numpy_to_vtk(data)
        array.SetName(name)
        self.data.GetCellData().AddArray(array)

    def has_point_array(self, name):
        return self.pointData.HasArray(name)

    def has_cell_array(self, name):
        return self.cellData.HasArray(name)

    def removePointData(self, *names):
        """
        Remove arrays to the PointData arrays

        :param list<str> names: Names of the arrays to remove
        """
        for name in names:
            if not self.pointData.HasArray(name):
                log.error(f"Array {name!r} could not be found")
            else:
                # log.debug(f"Removing PointDataArray {name!r}")
                self.pointData.RemoveArray(name)

    def removeCellData(self, *names):
        """
        Remove arrays to the CellData arrays

        :param list<str> names: Names of the arrays to remove
        """
        for name in names:
            if not self.cellData.HasArray(name):
                log.error(f"Array {name!r} could not be found")
            else:
                # log.debug(f"Removing CellDataArray {name!r}")
                self.cellData.RemoveArray(name)

    def clear(self):
        """ Removes all cell and point data arrays
        """
        for name in self.pointDataNames:
            self.removePointData(name)
        for name in self.cellDataNames:
            self.removeCellData(name)

    def hasPointData(self, name):
        return name in self.pointDataNames

    def hasCellData(self, name):
        return name in self.cellDataNames

    def getPointDataArray(self, name):
        """ Return the array of name `name`
        """
        if not self.hasPointData(name):
            msg = (
                f"{name!r} was not found in vtk PointData arrays: {self.pointDataNames}"
            )
            log.error(msg)
            raise RuntimeError(msg)
        return vtk_to_numpy(self.pointData.GetArray(name))

    def getCellScalars(self):
        """ Return the cellData scalars
        """
        return vtk_to_numpy(self.cellData.GetScalars())

    def getCellDataArray(self, name):
        """ Return the array of name `name`
        """
        if not self.hasCellData(name):
            msg = f"{name!r} was not found in vtk CellData arrays: {self.cellDataNames}"
            log.error(msg)
            raise RuntimeError(msg)
        return vtk_to_numpy(self.cellData.GetArray(name))

    @property
    def pointDataNames(self):
        """ Return the names of the point data arrays
        """
        names = list()
        for i in range(self.pointData.GetNumberOfArrays()):
            names.append(self.pointData.GetArrayName(i))
        return names

    @property
    def cellDataNames(self):
        """ Return the names of the cell data arrays
        """
        names = list()
        for i in range(self.cellData.GetNumberOfArrays()):
            names.append(self.cellData.GetArrayName(i))
        return names

    @property
    def pointData(self):
        return self.data.GetPointData()

    @property
    def cellData(self):
        return self.data.GetCellData()

    @vtkFilter
    def apply_filter(self, filter_name, **kwargs):
        """
        Explicit call to a vtk filter.
        Filter can be called without the `vtk`, passing options as `kwargs`.
        """
        return filter_name

    @vtkFilter
    def PointDataToCellData(self, SetPassPointData=True, **kwargs):
        return

    def CustomPointDataToCellData(self, point_name, new_name=None, method=np.max):
        """
        Computes conversion between point to cell data by only considering
        the maximum value at cell vertices
        """
        if new_name is None:
            new_name = point_name

        pt_data = self.getPointDataArray(point_name)
        new_cell_data = np.empty(self.nbCells, dtype=np.uint8)
        for i, cell in enumerate(self.cells):
            vals = [pt_data[p_id] for p_id in cell]
            new_cell_data[i] = method(np.array(vals))

        self.addCellData(new_cell_data, new_name)

    def CustomCellDataToPointData(
        self, cell_name=None, new_name=None, method=np.max, scalars=False
    ):
        """
        Computes conversion between cell to point data by only considering
        the maximum value at cell vertices
        """
        if self.cellsAroundPoints is None:
            self.buildCellsAroundPoints()

        if new_name is None:
            new_name = cell_name

        if scalars:
            cell_data = self.getCellScalars()
        else:
            cell_data = self.getCellDataArray(cell_name)
        new_point_data = np.empty(self.nbPoints, dtype=np.uint8)

        for p_id in range(self.nbPoints):
            neighbors = self.cellsAroundPoints[p_id]
            vals = [cell_data[c_id] for c_id in neighbors]
            new_point_data[p_id] = method(np.array(vals))

        self.addPointData(new_point_data, new_name)

    @vtkFilter
    def CellDataToPointData(self, SetPassCellData=True, **kwargs):
        return

    @vtkFilter
    def FillHolesFilter(self, SetHoleSize=100000, **kwargs):
        return

    def convertToPolyData(self, clean=True):
        self.GeometryFilter()
        if clean:
            self.CleanPolyData()

    def clean(self):
        self.CleanPolyData()

    def getVolume(self):
        """
        Compute volume of vtkPolyData

        .. warning::
            You may need to close the surface with `vtkFillHolesFilter`
        """
        filter = vtk.vtkMassProperties()
        filter.SetInputData(self.data)
        filter.Update()
        return filter.GetVolume()

    def getCenter(self):
        """
        Computes the barycenter. Weights are supported by vtk but not implemented here
        """
        filter = vtk.vtkCenterOfMass()
        filter.SetInputData(self.data)
        filter.Update()
        return filter.GetCenter()

    def intersect(self, point_1, point_2):
        """
        Computes the intersection of the mesh with the line defined by two points.

        :param point_1: First point of the line (list or np.array, of size 3)
        :param point_2: Second point of the line (list or np.array, of size 3)

        :return: Coordinates of n intersection points
        """
        tree = vtk.vtkOBBTree()
        tree.SetDataSet(self.data)
        tree.BuildLocator()
        intersectPoints = Points()
        tree.IntersectWithLine(point_1, point_2, intersectPoints, None)
        # log.debug(f"Intersection numPoints: {len(intersectPoints)}")
        return intersectPoints.as_np()

    def buildPointLocator(self):
        self.pointLocator = vtk.vtkPointLocator()
        self.pointLocator.SetDataSet(self.data)
        self.pointLocator.BuildLocator()

    def closestPoint(self, point):
        """
        Return the id and coordinates of the closest point to `point` in the mesh

        :param point: Coordinates of a point

        :return: Index and coordinates of nearest mesh point
        """
        if self.pointLocator is None:
            self.buildPointLocator()
        idx = self.pointLocator.FindClosestPoint(point)
        return idx, self.data.GetPoint(idx)

    def transform(self, transform=None, **kwargs):
        """
        Shorthand method to transform the mesh.
        See <gaetools.meshing.prefabs.BuildTransform> for arguments.
        """
        if transform is None:
            transform = BuildTransform(**kwargs)

        if self.type is vtk.vtkPolyData:
            self.TransformPolyDataFilter(SetTransform=transform)
        else:
            self.TransformFilter(SetTransform=transform)

        return transform

    def dilate(self, scale):
        origin = np.array(self.BBCenter)
        self.transform(translate=-origin)
        self.transform(scale=scale)
        self.transform(translate=origin)

    def interpolate_to_image(
        self, point_array_name, spacing=np.ones(3), name_valid_mask="vtkValidPointMask"
    ):
        log.warning(
            f"Weird behavior, checked that array_passed is really {point_array_name!r}"
        )
        # array = self.getPointDataArray(point_array_name)
        # self.clear()
        # self.addPointData(array, point_array_name)
        log.debug(f"Interpolate from {self.pointDataNames[0]!r}")

        bounds = self.bounds
        spacing = np.asarray(spacing, np.float64)
        dim = np.array(
            [
                np.ceil((bounds[2 * i + 1] - bounds[2 * i]) / spacing[i])
                for i in range(3)
            ],
            np.int,
        )
        origin = np.array([bounds[2 * i] + spacing[i] / 2 for i in range(3)])

        log.debug(f"{dim[0] * dim[1] * dim[2]} pixels in image")

        img_template = vtk.vtkImageData()
        img_template.SetExtent(0, dim[0] - 1, 0, dim[1] - 1, 0, dim[2] - 1)
        img_template.SetSpacing(spacing)
        img_template.SetDimensions(dim)
        img_template.SetOrigin(origin)

        obj = Object(img_template)
        obj.ProbeFilter(
            SetSourceData=self.data, SetValidPointMaskArrayName=name_valid_mask
        )

        # pts = (
        #     probe_filter.GetValidPoints()
        # )  # Get the list of point ids in the output that contain attribute data interpolated from the source.
        # log.debug(f"nbPts interpolated (valid mask): {pts.GetSize()}")

        return obj

    def getBoundingBox(self):
        """
        Computes the bounding box
        """
        box = vtk.vtkBoundingBox()
        for p in self.points:
            box.AddPoint(p)
        return box

    @property
    def BBDiag(self):
        return self.getBoundingBox().GetDiagonalLength()

    def getBBCenter(self):
        center = np.empty(3)
        self.getBoundingBox().GetCenter(center)
        return center

    @property
    def BBCenter(self):
        return self.getBBCenter()

    @deprecated("Avoid using it")
    def buildTrianglesOnBoundary(self, **array_names):
        """
        Creates a triangulation for points on boundary
        """
        # Convert ugrid to triangle mesh (aka polydata)
        polydata = self.copy()
        polydata.convertToPolyData()

        # Instanciate the point locator on ugrid
        pointLocator = vtk.vtkPointLocator()
        pointLocator.SetDataSet(self.data)
        pointLocator.BuildLocator()

        # Get cells and types
        cells = self.data.GetCells()
        cells_type = self.cellTypes

        for cell in polydata.cells:
            # For each triangle on polydata, find the indices of the corresponding vertices on ugrid
            new_cell = Cell(
                ids=[pointLocator.FindClosestPoint(coord) for coord in cell.points]
            )
            cells.InsertNextCell(new_cell)
            cells_type = np.append(cells_type, new_cell.type.val)

        # Set the updated cells array
        # self.data.SetCells(cells_type, cells)
        self.cells = cells

        # Add value in cell arrays to complete the added cells, see the kwargs
        for name in self.cellDataNames:
            if not name in array_names:
                log.warning(
                    f"{name!r} is not touched, file may be corrupted, pass '{name}=value' to add dumb values to cell array"
                )
            else:
                arr = self.getCellDataArray(name)
                shp = arr.shape[-1] if arr.ndim > 1 else 1
                temp_full = np.repeat(
                    array_names.pop(name, np.zeros(shp)), polydata.nbCells
                )
                new_array = np.concatenate((arr, temp_full))
                self.removeCellData(name)
                self.addCellData(new_array, name)

        for unused in array_names:
            log.warning(
                f"{unused!r} does not seem to be used, check that it exists in array"
            )

    def neighbors(self, pointId=None, cellId=None):
        if pointId is not None:
            if self.cellsAroundPoints is None:
                self.buildCellsAroundPoints()

            return self.cellsAroundPoints[pointId]

        if cellId is not None:
            raise NotImplementedError

    def buildCellsAroundPoints(self):
        cellsAroundPoints = [[] for _ in range(self.nbPoints)]
        for cell_id, cell in enumerate(self.cells):
            for pid in cell:
                cellsAroundPoints[pid].append(cell_id)
        self.cellsAroundPoints = np.array(cellsAroundPoints)

    def buildTriangles(self, newMesh=False, newCellData=None):
        mesh = self.copy()
        # mesh = Object(self.copy().data)
        mesh.clear()
        mesh.filterCellsTo(10)

        triangles = []
        for t in mesh.cells:
            for i in range(4):
                triangles.append(sorted([t[i], t[(i + 1) % 4], t[(i + 2) % 4]]))

        triangles = np.unique(np.array(triangles), axis=0)

        if newMesh:
            new_mesh = Object(type(self.data)())
            new_mesh.points = mesh.points
        else:
            new_mesh = mesh

        dtype = self.types.TRIANGLE
        cell = dtype.new()
        cells = new_mesh.data.GetCells()
        cellTypes = new_mesh.cellTypes

        for t in triangles:
            for i, point_id in enumerate(t):
                cell[i] = point_id
            cells.InsertNextCell(cell.data)

        cellTypes = np.concatenate(
            (cellTypes, np.ones(triangles.shape[0], dtype=np.uint8) * dtype.val)
        )
        new_mesh.data.SetCells(cellTypes, cells)

        # TODO Debug here
        # # Mesh data
        # if newCellData is not None:
        #     new_cell_data = np.ones(triangles.shape[0]) * newCellData
        #     for c in self.cellDataNames:
        #         arr = np.concatenate((self.getCellDataArray(c), new_cell_data))
        #         new_mesh.addCellData(arr, c)
        #
        #     for p in self.pointDataNames:
        #         new_mesh.addPointData(self.getPointDataArray(p), p)

        return new_mesh

    @timer
    def make_manifold(self, verbose=False, inPlace=True):
        """Reconstruct a manifold clean surface from input mesh.

        Requires pymeshfix

        Parameters
        ----------
        verbose : bool, optional
            Controls output printing.  Default False.
        inPlace : bool, optional
            Update mesh in place or return new instance
        """
        try:
            import pymeshfix
        except ImportError:
            raise ImportError(
                "pymeshfix not installed.  Please run:\n" "pip install pymeshfix"
            )

        if self.type is vtk.vtkUnstructuredGrid:
            log.warning(
                "'make_manifold' method only accepts polydata as input, converting now"
            )
            self.convertToPolyData()

        # Run meshfix
        meshfix = pymeshfix.MeshFix(self.points.as_np(), self.cells.as_np())
        if not verbose:
            # TODO fix __enter__ method
            # with suppress_stdout_stderr:
            meshfix.repair()
        else:
            meshfix.repair(verbose)

        # New <Object> containing the manifold surface
        repaired_mesh = Object.new("pd")
        # overwrite this object with cleaned mesh
        repaired_mesh.points = meshfix.v
        repaired_mesh.cells = meshfix.f

        # Interpolate data arrays
        repaired_mesh.interpolate(self.copy())

        # update or return
        if inPlace:
            self.data = repaired_mesh.data
        else:
            return repaired_mesh

    def custom_remove_duplicates(self):
        cells = self.cells.to_numpy()
        u = np.unique(np.sort(cells), axis=0)
        self.cells = u

    @property
    def pointIdsAroundPoint(self):
        if self._pointIdsAroundPoint is None:
            self.buildPointIdsAroundPoint()
        return self._pointIdsAroundPoint

    def getPointsAroundPoint(self, point_id):
        return self.pointIdsAroundPoint[point_id]

    def buildPointIdsAroundPoint(self):
        ids_around_point = [[] for _ in range(self.nbPoints)]
        for c in self.cells:
            for p in c:
                ids_around_point[p].extend(c.ids)

        for p in range(self.nbPoints):
            pp = list(dict.fromkeys(ids_around_point[p]))
            pp.remove(p)
            ids_around_point[p] = pp

        self._pointIdsAroundPoint = ids_around_point
        return ids_around_point

    @unstructured_grid_method
    def buildTetrasAroundPoint(self):
        self.tetras_around_points = [[] for _ in range(self.nbPoints)]
        for cell in self.tetras:
            for pid in cell:
                self.tetras_around_points[pid].append(cell.id)

    @unstructured_grid_method
    def tetrasAroundPoint(self, p_id):
        if self.tetras_around_points is None:
            self.buildTetrasAroundPoint()
        return self.tetras_around_points[p_id]

    @polydata_method
    def smooth(self, nb_iter=20, relax=0.1):
        self.SmoothPolyDataFilter(
            SetNumberOfIterations=nb_iter, SetRelaxationFactor=relax
        )

    def surface_geodesic(self, array_name=None, set_start_coord=None, **kwargs):
        """
        :param kwargs: See https://vtk.org/doc/nightly/html/classvtkDijkstraGraphGeodesicPath.html
        ex: SetStartVertex, SetEndVertex, StopWhenEndReachedOn
        """
        if self.type is not vtk.vtkPolyData:
            log.warning(
                f"{Object.surface_geodesic} is intended for PolyData only, converting to polydata"
            )
            self.convertToPolyData()

        if set_start_coord is not None:
            i, _ = self.closestPoint(set_start_coord)
            kwargs.update({"SetStartVertex": i})

        filter = vtk.vtkDijkstraGraphGeodesicPath()
        filter.SetInputData(self.data)

        for k, v in kwargs.items():
            if not isinstance(v, list):
                v = [v]
            getattr(filter, k)(*v)

        filter.Update()

        distance_map = vtk.vtkDoubleArray()
        filter.GetCumulativeWeights(distance_map)
        distance_map = vtk_to_numpy(distance_map)

        if array_name is not None:
            self.addPointData(distance_map, array_name)

        return distance_map

    def getVoronoi(
        self, seed_points, mesh=None, method=None, array_name=None, recenter=False
    ):
        from MeshObject.voronoi import Voronoi

        if mesh is None:
            mesh = self.copy()

        v = Voronoi(mesh=mesh, seed_points=seed_points)
        array = v.run(method)

        if recenter:
            v.recenter_points()

        if array_name is not None:
            self.addPointData(array, array_name)

        return v

    def getVoronoi_Lloyd(self, nbPoints, mesh=None):
        """
        Returns the ids of the voronoi centers on the mesh
        """
        from MeshObject.voronoi import PointCloudUtils

        if mesh is None:
            mesh = self.copy()

        v = PointCloudUtils(mesh)
        seeds_id = v.Llyod(nbPoints)
        return seeds_id

    def threshold(self, btw, array_name, method=None, to_polydata=True):
        if method is None:
            method = "point"

        # if method is not "point" or method is not "cell":
        #     raise NotImplemented

        self.Threshold(
            ThresholdBetween=[btw[0], btw[1]],
            SetInputArrayToProcess=[
                0,
                0,
                0,
                getattr(vtk.vtkDataObject, f"FIELD_ASSOCIATION_{method.upper()}S"),
                array_name,
            ],
        )
        if self.nbPoints == 0:
            log.warning("Threshold may have failed, mesh has 0 points")

        if to_polydata:
            self.convertToPolyData()

    @property
    def point_normals(self):
        normals_filter = self.copy()
        normals_filter.TriangleMeshPointNormals()
        return normals_filter.getPointDataArray("Normals")

    @polydata_method
    @property
    def cell_normals(self):
        normals_filter = self.copy()
        normals_filter.PolyDataNormals()
        return normals_filter.getPointDataArray("Normals")

    @timer
    @unstructured_grid_method
    def addTriangles(self, replace=False, val_added_cells=0):
        """
        Builds all triangles in the unstructured mesh
        :param replace: stack cells (default) or replace cells array
        :param val_added_cells: if replace is False, default value for the new cells in cell data arrays
        """
        log.info("Building triangles")
        self.filterCellsTo(Object.types.TETRA)

        tri, tri_sorted = [], []
        for cell in self.cells:
            c = cell.as_np()
            for j in range(4):
                a, b, d = c[j % 4], c[(j + 1) % 4], c[(j + 2) % 4]
                tri.append([a, b, d])
                tri_sorted.append(sorted([a, b, d]))
        #         tri.append(sorted([c[j % 4], c[(j + 1) % 4], c[(j + 2) % 4]]))
        # tri = np.unique(np.array(tri), axis=0)
        # Global convention for nodes indexing is not lost, sorted triangle vertices indexes are kept apart from non sorted ones
        _, indexes = np.unique(np.array(tri_sorted), axis=0, return_index=True)
        tri = np.array(tri)[indexes]

        if replace:
            self.cells = tri
        else:
            self.addCells(tri)
            ept = np.ones(tri.shape[0]) * val_added_cells
            for cd in self.cellDataNames:
                new_arr = np.concatenate((self.getCellDataArray(cd), ept))
                self.addCellData(new_arr, cd)

    # @unstructured_grid_method
    # def cleanUGrid(self):
    #     import paraview.simple as psimple
    #     import paraview.servermanager as pserver
    #
    #     tp = psimple.TrivialProducer()
    #     tp.GetClientSideObject().SetOutput(self.data)
    #
    #     psimple.CleantoGrid(tp)
    #     psimple.CleanCellstoGrid(tp)
    #
    #     self.data = pserver.Fetch(tp)

    @property
    def mmg_options(self, **kwargs):
        return MMGOptions(mesh=self, **kwargs)

    @polydata_method
    def tetrahedralize(
        self, mmg_opt: [List, MMGOptions] = None, inplace=True, **kwargs
    ):
        """
        Remeshing with mmg. Check in particular `-hausd` and `-hgrad` parameters.
        See more in the Tetrahedralize class.
        """
        from MeshObject.poly_meshing import Tetrahedralize

        kwargs.update({"mmg_opt": mmg_opt})
        with tempfile.NamedTemporaryFile(suffix=".vtk") as mesh:
            Tetrahedralize(self.data, mesh.name, **kwargs)
            obj = Object.read(mesh.name, type="ugrid")
        if inplace:
            self.data = obj.data
        else:
            return obj

    @polydata_method
    def triangulate(self, mmg_opt: [List, MMGOptions] = None, inplace=True, **kwargs):
        from MeshObject.poly_meshing import Triangulate

        kwargs.update({"mmg_opt": mmg_opt})
        with tempfile.NamedTemporaryFile(suffix=".vtk") as mesh:
            Triangulate(self.data, mesh.name, **kwargs)
            obj = Object.read(mesh.name, type="pd")
        if inplace:
            self.data = obj.data
        else:
            return obj

    @staticmethod
    def replace_polygons(filename):
        Convert(filename)

    @staticmethod
    def change_vtk_version(filename, v):
        with open(filename, "r") as f:
            content = f.read().split("\n")
        content[0] = f"# vtk DataFile Version {v}"
        with open(filename, "w") as f:
            f.write("\n".join(content))

    @classmethod
    def marching_cubes(
        cls, image, isovalue=0, smooth=True, make_manifold=False, manifold_verbose=False
    ):
        """
        Takes a <cardiac_utils.infoarray.Image> image and returns an Object<vtkPolyData>
        """
        try:
            from skimage import measure
        except ImportError:
            raise ImportError(
                "You need to install skimage to run the marching cubes method"
            )

        ### Deal with image touching border of the bounding box
        # 1. Add a thin layer (padding) of 1 voxel
        # origin = image.origin - [1, 1, 1]
        # image = np.pad(image, 1)
        # 2. Do nothing and fill holes later
        origin = image.origin

        vertices, faces, normals, values = measure.marching_cubes_lewiner(
            image.transpose((2, 1, 0)), level=isovalue
        )

        polydata = Object.from_elements(vertices, faces, dtype="pd")
        polydata.transform(translate=origin)
        polydata.FillHolesFilter()

        # values[values == 2] = 0  # Match legend color for further use in Pipeline
        # polydata.addPointData(values, "Ventricles")
        polydata.addPointData(values, "Values")
        polydata.addPointData(normals, "Normals")

        if make_manifold:
            # log.info("Start repairing (approx. 7min)")
            polydata.make_manifold(verbose=manifold_verbose)
        else:
            pass
            # log.info("No manifold check !")

        if smooth:
            polydata.smooth(nb_iter=40, relax=0.1)

        return polydata

    def to_image(self, spacing=None, filename=None, mode=None):
        """
        See <cardiac_utils.mesh.SpatialData.to_image(
            spacing, filename=None, inval=1, outval=0, bounds=None
        )>
        Return: <cardiac_utils.infoarray.Image> image

        mode: np.bool by default?
        """
        if not HAS_CARDIAC_UTILS:
            raise ImportError("The cardiac_utils package has not been found")

        if spacing is None:
            spacing = (1, 1, 1)

        sph = SpatialData(vtk_polydata=self.data)
        img = sph.to_image(spacing, filename=filename, bounds=self.bounds)

        if mode is None:
            return img
        elif mode == "nan":
            imc = img.astype(np.float64)
            imc[np.where(imc == 0)] = np.nan
            return imc
        elif mode == "-1":
            imc = img.astype(np.float64)
            imc[np.where(imc == 0)] = -1
            return imc

    def get_boundary_edges(self):
        from MeshObject.boundary_edges import getBoundaryEdges

        return getBoundaryEdges(self)


#### TEMPORARY
# Intended to be moved to a stand alone package when I will have time


import math


class Convert:
    def __init__(self, inp, out=None, copy_lines=False):
        if out is None:
            out = inp
        self.inp = inp
        self.out = out

        self.copy_lines = copy_lines

        self.read()
        self.replace()
        self.write()

    def read(self):
        with open(self.inp, "r") as f:
            content = f.read()
        self.original = content.split("\n")
        lines_original = list(map(lambda x: x.strip().split(" "), self.original))

        self.lines = None
        self.polys = None
        self.cells = None

        for i, l in enumerate(self.original):
            if "LINES" in l:
                self.lines = NewContent("LINES", lines_original, i)
            elif "POLYGONS" in l:
                self.polys = NewContent("POLYGONS", lines_original, i)
            elif "CELLS" in l:
                self.cells = NewContent("CELLS", lines_original, i)

    def replace(self):
        if self.polys is not None:
            self.original = self.polys.replace(self.original)
        if self.cells is not None:
            self.original = self.cells.replace(self.original)
        if self.lines is not None:
            self.original = self.lines.replace(self.original, replace=self.copy_lines)

    def write(self):
        self.original[0] = "# vtk DataFile Version 4.2"

        with open(self.out, "w") as f:
            f.write("\n".join(self.original))


class NewContent:
    def __init__(self, kw, content, ln):
        self.kw = kw
        self.ln = ln
        self.name = content[ln][0]
        self.no = int(content[ln][1])
        self.nc = int(content[ln][2])

        flat_list = [item for line in content[ln + 2 :] for item in line]
        flat_list = list(filter(("").__ne__, flat_list))

        self.offsets = list(map(int, flat_list[0 : self.no]))
        self.connectivity = list(
            map(int, flat_list[self.no + 2 : self.no + 2 + self.nc])
        )

        # print(self.offsets)
        # print(self.connectivity)

    @property
    def remove(self):
        return (
            self.ln,
            self.ln + math.ceil(self.no / 9) + math.ceil(self.nc / 9) + 3,
        )  # Or 4?

    def replace(self, lines, replace=True):
        nb_cells = self.no - 1
        if replace:
            new_content = [f"{self.kw} {nb_cells} {nb_cells + self.nc}"]
            for i in range(nb_cells):
                nb_points = self.offsets[i + 1] - self.offsets[i]
                ids = self.connectivity[self.offsets[i] : self.offsets[i + 1]]
                new_content.append(f"{nb_points} {' '.join(map(str, ids))}")
        else:
            new_content = []

        lines_to_keep = lines
        a, b = self.remove
        del lines_to_keep[a:b]
        lines_to_keep[a:a] = new_content
        lines_to_keep = list(filter(("").__ne__, lines_to_keep))
        return lines_to_keep


########## END TEMPORARY


log = logging.getLogger(__name__)
